/*
 *
 * Hierarchical Penrose Tiling implementation.
 *
 * see details for:
 *
 * Victor Ostromoukhov, Charles Donohue, Pierre-Marc Jodoin,
 * Fast Hierarchical Importance Sampling with Blue Noise Properties,
 * ACM SIGGRAPH 2004.
 * http://www.iro.umontreal.ca/~ostrom/publications/
 *
 * How to compile:
 *
 * o for Mac OS X
 *
 * $ gcc penrose.c -framework OpenGL -framework GLUT -lobjc
 *
 * o for Windows(cygwin)
 *
 * $ gcc penrose.c -lopengl32 -lglu32 -lglut32
 *
 * Syoyo FUJITA syoyo@users.sourceforge.net
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
//#include <assert.h>

#include <sys/time.h>

#if defined(__APPLE__) && defined(__MACH__)
#include <OpenGL/OpenGL.h>
#include <OpenGL/CGLCurrent.h>
#include <OpenGL/gl.h>
#include <OpenGL/glext.h>
#include <GLUT/glut.h>

#define RECTANGLER_TARGET GL_TEXTURE_RECTANGLE_EXT
#else
#include <GL/glut.h>

// sorry! only for NV card
/* NV_texture_rectangle */
#define GL_NV_texture_rectangle           1
#define GL_TEXTURE_RECTANGLE_NV           0x84F5
#define GL_TEXTURE_BINDING_RECTANGLE_NV   0x84F6
#define GL_PROXY_TEXTURE_RECTANGLE_NV     0x84F7
#define GL_MAX_RECTANGLE_TEXTURE_SIZE_NV  0x84F8

#define RECTANGLER_TARGET GL_TEXTURE_RECTANGLE_NV 
#endif

#include "rgbe.h"
#include "glerror.h"

/* There are 6 different tiles in total */
#define TILE_A 0
#define TILE_B 1
#define TILE_C 2
#define TILE_D 3
#define TILE_E 4
#define TILE_F 5

#define GOLDEN_RATIO   1.61803			/* GR = (1 + sqrt(5)) / 2 */
#define GOLDEN_RATIO_1 1.0/GOLDEN_RATIO			/* GR^(-1)	*/
#define GOLDEN_RATIO_2 1.0/(GOLDEN_RATIO*GOLDEN_RATIO)	/* GR^(-2)	*/
/* vertical height of tile C and D = sqrt(1*1 - ((GD^-1)/2)^2) */
#define HEIGHT_C 0.95105651629515353
/* vertical height of tile E and F = sqrt(1*1 - (GD/2)^2) */
#define HEIGHT_E 0.58778525229247314

#define MAX_SUBD 8
#define NTILES 4

/* CIE color definitions */
#define CIE_rf 0.265074126
#define CIE_gf 0.670114631
#define CIE_bf 0.064811243

#define CIEcol(v) (CIE_rf * (v)[0] + CIE_gf * (v)[1] + CIE_bf * (v)[2])

typedef struct _tile_t
{
	//double          pos[2];	/* center position 		*/
	double          v[3][2];	/* tile's vertex position	*/
	double          dir[2][2];	/* tile's direction		*/
	double          rot;		/* absolute 2D rotation angle.	*/  
	int             level;		/* subdivision level		*/
	int             scale;		/* scale of tile		*/
	int             type;		/* type of this tile		*/
	unsigned char   Fcode[2 * MAX_SUBD]; /* Fibonacci codes		*/
	struct _tile_t *child[4];	/* child tile(s)		*/
	struct _tile_t *parent;
} tile_t;

float colors[6][3] = {
	{ 1.0, 0.0, 1.0 },
	{ 0.0, 1.0, 0.0 },
	{ 0.6, 0.7, 1.0 },
	{ 0.0, 0.0, 1.0 },
	{ 1.0, 0.8, 0.6 },
	{ 1.0, 0.5, 0.3 }
};

double vertices[6] = {
	-GOLDEN_RATIO / 2.0, 0.0,
	 GOLDEN_RATIO / 2.0, 0.0,
	                0.0, HEIGHT_E			/* tile E and F */
};

#define WINDOW_WIDTH  800
#define WINDOW_HEIGHT 400 

#define MAPSIZE 500
#define MAX_SAMPLES MAPSIZE * MAPSIZE

static int      width, height;

static tile_t *gtiles[NTILES*NTILES*4];
static int     showlevel = 1;
static int     FibNum[MAX_SUBD * 2 + 3];	/* fibonacci number */
static float   gscale = MAPSIZE/2;
static float   gimportancemap[MAPSIZE/2][MAPSIZE];
static double  gmag = 10.0;

static double  gsamples[MAX_SAMPLES][2];
static double  gsamplecvs[MAX_SAMPLES][2];
static int     gnsamples;
static GLuint  gtexid;
static int     show_tiles = 1;
static int     show_hierarcy = 0;

void           draw_tile(tile_t *tile, int level);
void           draw_tile_notraverse(tile_t *tile);

/* table for corrective vectors */
static float cv[8][21][2] =
{{{0.014565, 0.0183169}, {-0.0142356, -0.00758454}, {-0.010021, -0.341873}, {-0.0131048, -0.19669}, 
  {-0.0354038, 0.046953}, {0.230055, 0.391294}, {-0.16039, 0.446062}, {-0.227107, 0.367212}, {0.00198984, 1.0218}, 
  {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}}, 
 {{0.0104519, -0.0651952}, {0.198966, -0.222273}, {-0.0474862, -0.568896}, {-0.0134488, -0.25001}, 
  {-0.13474, -0.059871}, {0.466686, 0.121227}, {-0.455974, -0.0690133}, {-0.478023, -0.0058658}, 
  {0.0657471, 0.843717}, {0.256898, 0.295477}, {0.000114695, 0.00810291}, {0.0148141, 0.240328}, 
  {-0.0231445, 0.0765225}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}}, 
 {{-0.00522375, -0.0714328}, {0.332176, -0.265696}, {-0.0814839, -0.611528}, {0.00652924, -0.13296}, 
  {-0.185437, 0.0254077}, {0.450314, 0.025685}, {-0.458329, -0.215996}, {-0.551482, -0.237815}, 
  {0.102447, 0.728033}, {0.286354, 0.316227}, {-0.000659441, 0.028732}, {0.0152405, 0.512455}, 
  {-0.0683259, 0.213807}, {0.0266667, 0.00988119}, {0.019467, 0.0113397}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}},
 {{-0.0244669, -0.0515528}, {0.394376, -0.301079}, {-0.12632, -0.591921}, {0.0664285, 0.108731}, 
  {-0.135581, 0.178174}, {0.352024, -0.0398514}, {-0.471673, -0.320062}, {-0.625742, -0.410088}, 
  {0.158362, 0.558113}, {0.297232, 0.293498}, {-0.00459988, 0.074168}, {-0.0336619, 0.696709}, 
  {-0.170609, 0.494946}, {0.0814721, 0.0328496}, {0.0608224, 0.0378825}, {0.0191699, 0.0233764}, 
  {-0.00366271, 0.0144315}, {0, 0}, {0, 0}, {0, 0}, {0, 0}}, 
 {{-0.0575956, -0.0272753}, {0.366739, -0.299236}, {-0.203983, -0.511355}, {0.174012, 0.268644}, 
  {0.00279778, 0.334415}, {0.18246, -0.00227333}, {-0.469253, -0.316594}, {-0.68479, -0.462175}, 
  {0.195021, 0.391098}, {0.257835, 0.197128}, {-0.00451144, 0.098079}, {-0.0880223, 0.723411}, 
  {-0.238336, 0.609767}, {0.18772, 0.0766329}, {0.141226, 0.089601}, {0.0565302, 0.0692062}, 
  {-0.0119856, 0.0487706}, {-0.0185962, 0.0203027}, {0.00100156, 0.0101348}, {0, 0}, {0, 0}}, 
 {{-0.0886771, -0.0147547}, {0.269449, -0.303973}, {-0.233732, -0.388322}, {0.235346, 0.338068}, 
  {0.158806, 0.373381}, {0.00874489, 0.141831}, {-0.478748, -0.238658}, {-0.688808, -0.385529}, 
  {0.242988, 0.282458}, {0.217421, 0.106127}, {-0.00449596, 0.0796765}, {-0.122395, 0.598062}, 
  {-0.272818, 0.576112}, {0.234976, 0.105758}, {0.183237, 0.12421}, {0.129803, 0.158582}, {-0.0272725, 0.120763}, 
  {-0.0524825, 0.0645941}, {-0.00630879, 0.0305341}, {-0.0155945, 0.0175939}, {-0.0127748, 0.0322832}}, 
 {{-0.0918686, 0.00448731}, {0.167911, -0.277722}, {-0.207131, -0.267332}, {0.22313, 0.316075}, 
  {0.200598, 0.360563}, {-0.0713932, 0.303626}, {-0.414882, -0.0852619}, {-0.604657, -0.212697}, 
  {0.240506, 0.275773}, {0.177329, 0.0440231}, {0.00101409, 0.0163503}, {-0.116679, 0.35887}, {-0.231965, 0.37873}, 
  {0.238258, 0.110352}, {0.189403, 0.126668}, {0.154135, 0.187339}, {-0.0321777, 0.177625}, {-0.10619, 0.150575}, 
  {-0.0341313, 0.0657763}, {-0.0380597, 0.0465556}, {-0.0355876, 0.0761992}}, 
 {{-0.0703082, 0.0336306}, {0.100832, -0.254953}, {-0.130627, -0.171264}, {0.149742, 0.243892}, 
  {0.153357, 0.283324}, {-0.0570642, 0.443626}, {-0.299722, 0.0913463}, {-0.48891, -0.0415333}, 
  {0.215232, 0.313909}, {0.144138, 0.00860654}, {0.00379, -0.0302774}, {-0.0843071, 0.201935}, 
  {-0.195705, 0.227506}, {0.211236, 0.106754}, {0.174136, 0.113369}, {0.157354, 0.188698}, {-0.0282292, 0.213901}, 
  {-0.112004, 0.19926}, {-0.0825244, 0.0786043}, {-0.084577, 0.103457}, {-0.0790835, 0.169332}}};

void
normalize(double v[2])
{
	double len;

	len = sqrt(v[0] * v[0] + v[1] * v[1]);
	if (len >= 1.0e-6) {
		len = 1.0 / len;
		v[0] *= len;
		v[1] *= len;
	}
}

void
normalize3(double v[3])
{
	double len;

	len = sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2]);
	if (len >= 1.0e-6) {
		len = 1.0 / len;
		v[0] *= len;
		v[1] *= len;
		v[2] *= len;
	}

}

/* calculates tile's direction from tile.v[1..3] */
void
calc_tile_dir(tile_t *tile)
{
	/* x coord dir */
	tile->dir[0][0] = tile->v[1][0] - tile->v[0][0];	
	tile->dir[0][1] = tile->v[1][1] - tile->v[0][1];	

	normalize(tile->dir[0]);

	/* y coord dir */
	tile->dir[1][0] = -tile->dir[0][1];
	tile->dir[1][1] = tile->dir[0][0];

	//printf("dirx = %f, %f, diry = %f, %f\n",
	//	tile->dir[0][0], tile->dir[0][1],
	//	tile->dir[1][0], tile->dir[1][1]);
}

void
swap_dir(double d1[2], double d2[2])
{
	double tmp;

	tmp = d2[0]; d2[0] = d1[0]; d1[0] = tmp;
	tmp = d2[1]; d2[1] = d1[1]; d1[1] = tmp;
}

static tile_t *
new_tile()
{
	tile_t *p;

	p = (tile_t *)malloc(sizeof(tile_t));

	p->child[0] = NULL;
	p->child[1] = NULL;
	p->child[2] = NULL;
	p->child[3] = NULL;
	p->parent   = NULL;
	p->dir[0][0] = 1.0;
	p->dir[0][1] = 0.0;
	p->dir[1][0] = 0.0;
	p->dir[1][1] = 1.0;
	memset(p->Fcode, 0, 2 * MAX_SUBD);

	

	return p;
}

/* convert F-Code to decimal number representation. */
int
FibToDecimal(unsigned char Fcode[], int len)
{
	int i;
	int accumulator = 0;
	int maxlen;

	maxlen = len - 1;
	if (maxlen >= (MAX_SUBD * 2)) maxlen = MAX_SUBD * 2;

	for (i = 0; i < maxlen; i++) {
		accumulator += Fcode[i] * FibNum[i + 2];
	}

	//printf("fib2dec = %d\n", accumulator);

	return accumulator;
}

/* function Psi(x) that maps a real positive x onto interval [0..1],
 * as shown in Figure 12
 */
double
Psi(double x)
{
	double p;

	if (x < 1.0e-6) return 0.0;

	/* Psi = log_{GR^2}(sqrt(5) * x) */
	p = log(sqrt(5) * x) / log(GOLDEN_RATIO * GOLDEN_RATIO);

	p = p - floor(p);	// p mod 1

	return p;
}

void
print_Fcode(tile_t *tile)
{
	int i;

	printf("tile = %d, level = %d, Fcode = ", tile->type, tile->level); 
	for (i = (tile->level - 1) * 2 - 1; i >= 0; i--) {
		printf("%d", tile->Fcode[i]);
	}
	printf(" , dec = %d\n", FibToDecimal(tile->Fcode, tile->level * 2));
	printf("\n");
}

int
get_structual_index(tile_t *tile)
{
	int is;
	int start;

	/* get structual index from first six bits of tile->Fcode */

	start = (tile->level - 1) * 2 - 6;
	if (start < 0) start = 0;

	is = FibToDecimal(&tile->Fcode[start], 7);

	//assert(is < 21);

	return is;
}

/*
 * Rasterize a triangle tile to find maximum inportance within it.
 */
float
FindMaxImportance(tile_t *t)
{
	/* TODO: optimize using integer arithmetric operation */

	/* simple polygon scan conversion routine. */
	int i, j;
	int ys, ye;
	double x[3], y[3], xm, xl, xr, ym;
	double tmp;
	float max_imp = 0.0;

	if (t->type == TILE_A || t->type == TILE_B) return 0.0;

	/* sort vertices by y coordinate */
	x[0] = t->v[0][0]; y[0] = t->v[0][1];
	x[1] = t->v[1][0]; y[1] = t->v[1][1];
	x[2] = t->v[2][0]; y[2] = t->v[2][1];

	/* simple O(n^2) sort */
	for (j = 0; j < 3; j++) {
		for (i = j + 1; i < 3; i++) {
			if (y[j] > y[i]) {
				tmp = x[j]; x[j] = x[i]; x[i] = tmp;
				tmp = y[j]; y[j] = y[i]; y[i] = tmp;
			}
		}
	}

	/* x coord of middle point */
	xm = x[2] + ((x[0] - x[2]) * (y[1] - y[2])) / (y[0] - y[2]);
	if (xm > x[1]) {
		tmp = xm; xm = x[1]; x[1] = tmp;
	}

	if (fabs(y[2] - y[1]) > 1.0e-3) {
		/* rasterize upper triangle */

		ys = (int)(y[2] - 0.5); ye = (int)y[1];
		if (ys < 0) ys = 0;
		if (ye < 0) ye = 0;
		if (ys >= MAPSIZE/2) ys = MAPSIZE/2-1;
		if (ye >= MAPSIZE/2) ye = MAPSIZE/2-1;

		for (i = ys; i >= ye; i--) {
			ym = (double)i;

			xl = x[2]+((xm - x[2]) * (ym - y[2])) / (y[1] - y[2]);
			xr = x[2]+((x[1] - x[2]) * (ym - y[2])) / (y[1] - y[2]);

			if (xl > xr) {
				tmp = xl; xl = xr; xr = tmp;
			}

			//assert(xl <= xr);

			if (xl < 0) xl = 0;
			if (xr < 0) xr = 0;
			if (xl >= MAPSIZE) xl = MAPSIZE-1;
			if (xr >= MAPSIZE) xr = MAPSIZE-1;

			for (j = (int)xl; j <= (int)xr; j++) {
				if (max_imp < gimportancemap[i][j]) {
					max_imp = gimportancemap[i][j];
				}
			}
		}
	}

	if (fabs(y[1] - y[0]) > 1.0e-3) {
		/* rasterize lower triangle */

		ys = (int)(y[1] - 0.5); ye = (int)y[0];
		if (ys < 0) ys = 0;
		if (ye < 0) ye = 0;
		if (ys >= MAPSIZE/2) ys = MAPSIZE/2-1;
		if (ye >= MAPSIZE/2) ye = MAPSIZE/2-1;

		for (i = ys; i >= ye; i--) {
			ym = (double)i;

			xr = x[1]+((xm - x[1]) * (ym - y[1])) / (y[0] - y[1]);
			xl = x[1]+((x[0] - x[1]) * (ym - y[1])) / (y[0] - y[1]);

			if (xl > xr) {
				tmp = xl; xl = xr; xr = tmp;
			}

			//assert(xl <= xr);

			if (xl < 0) xl = 0;
			if (xr < 0) xr = 0;
			if (xl >= MAPSIZE) xl = MAPSIZE-1;
			if (xr >= MAPSIZE) xr = MAPSIZE-1;

			for (j = (int)xl; j <= (int)xr; j++) {
				if (max_imp < gimportancemap[i][j]) {
					max_imp = gimportancemap[i][j];
				}
			}
		}

	}

	return max_imp;
}

/*
 * Subdivide tile by the production rules using equation(1) in the paper.
 */
static void
subdivide(tile_t *tile)
{
	tile_t *p;

	switch (tile->type) {
		case TILE_A:
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_B;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v0 */
			p->v[0][0] = tile->v[0][0];
			p->v[0][1] = tile->v[0][1];

			/* p.dir = tile.dir */
			memcpy(&p->dir[0][0], &tile->dir[0][0],
			       sizeof(double) * 4);

			/*  a* -> b00*  */
			memcpy(p->Fcode, tile->Fcode, 2 * MAX_SUBD);
			p->Fcode[2 * (tile->level - 1) + 0] = 0;
			p->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[0] = p;
		break;

		case TILE_B:
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_A;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v0 */
			p->v[0][0] = tile->v[0][0];
			p->v[0][1] = tile->v[0][1];

			/* p.dir = tile.dir + rot 180 */
			p->dir[0][0] = -tile->dir[0][0];
			p->dir[0][1] = -tile->dir[0][1];
			p->dir[1][0] = -tile->dir[1][0];
			p->dir[1][1] = -tile->dir[1][1];

			/*  b* -> a00*  */
			memcpy(p->Fcode, tile->Fcode, 2 * MAX_SUBD);
			p->Fcode[2 * (tile->level - 1) + 0] = 0;
			p->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[0] = p;
		break;

		case TILE_C:

			/* f */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_F;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v2 */
			p->v[0][0] = tile->v[2][0];
			p->v[0][1] = tile->v[2][1];

			/* p.v1  = tile.v0 */
			p->v[1][0] = tile->v[0][0];
			p->v[1][1] = tile->v[0][1];

			/* p.v2 = (1 - phi^-1) * tile.v2 + (tile.v1 * phi^-1) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[2][0]
				   +        GOLDEN_RATIO_1  * tile->v[1][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[2][1]
				   +        GOLDEN_RATIO_1  * tile->v[1][1];

			calc_tile_dir(p);
			tile->child[0] = p;
			
			/* c */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_C;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v1 */
			p->v[0][0] = tile->v[1][0];
			p->v[0][1] = tile->v[1][1];

			/* p.v1 = (1 - phi^-2) * tile.v1 + (tile.v2 * phi^-2) */
			p->v[1][0] = (1.0 - GOLDEN_RATIO_2) * tile->v[1][0]
				   +        GOLDEN_RATIO_2  * tile->v[2][0];
			p->v[1][1] = (1.0 - GOLDEN_RATIO_2) * tile->v[1][1]
				   +        GOLDEN_RATIO_2  * tile->v[2][1];

			/* p.v2 = tile.v0 */
			p->v[2][0] = tile->v[0][0];
			p->v[2][1] = tile->v[0][1];

			calc_tile_dir(p);

			tile->child[1] = p;

			/* a */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_A;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = (1 - phi^-1) * tile.v2 - (tile.v1 * phi^-1) */
			p->v[0][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[2][0]
				   +        GOLDEN_RATIO_1  * tile->v[1][0];
			p->v[0][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[2][1]
				   +        GOLDEN_RATIO_1  * tile->v[1][1];

			/* p.v1 = tile.v2 */
			p->v[1][0] = tile->v[2][0];
			p->v[1][1] = tile->v[2][1];

			calc_tile_dir(p);

			/* negate y, and swap x and y */
			p->dir[1][0] = -p->dir[1][0];
			p->dir[1][1] = -p->dir[1][1];
			swap_dir(p->dir[0], p->dir[1]);

			tile->child[2] = p;

			/*  c* -> { f00*, c10*, a10* } */
			memcpy(tile->child[0]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[1]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[2]->Fcode, tile->Fcode, 2*MAX_SUBD);

			tile->child[0]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[0]->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[1]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[1]->Fcode[2 * (tile->level - 1) + 1] = 1;

			tile->child[2]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[2]->Fcode[2 * (tile->level - 1) + 1] = 1;
		break;

		case TILE_D:
			
			/* e */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_E;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v1 */
			p->v[0][0] = tile->v[1][0];
			p->v[0][1] = tile->v[1][1];

			/* p.v1 = tile.v2 */
			p->v[1][0] = tile->v[2][0];
			p->v[1][1] = tile->v[2][1];

			/* p.v2 = (1 - phi^-2) * tile.v0 + (tile.v2 * phi^-2) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_2) * tile->v[0][0]
				   +        GOLDEN_RATIO_2  * tile->v[2][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_2) * tile->v[0][1]
				   +        GOLDEN_RATIO_2  * tile->v[2][1];

			calc_tile_dir(p);

			tile->child[0] = p;

			/* d */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_D;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = (1 - phi^-2) * tile.v0 + (tile.v2 * phi^-2) */
			p->v[0][0] = (1.0 - GOLDEN_RATIO_2) * tile->v[0][0]
				   +        GOLDEN_RATIO_2  * tile->v[2][0];
			p->v[0][1] = (1.0 - GOLDEN_RATIO_2) * tile->v[0][1]
				   +        GOLDEN_RATIO_2  * tile->v[2][1];

			/* p.v1 = tile.v0 */
			p->v[1][0] = tile->v[0][0];
			p->v[1][1] = tile->v[0][1];

			/* p.v2  = tile.v1 */
			p->v[2][0] = tile->v[1][0];
			p->v[2][1] = tile->v[1][1];

			calc_tile_dir(p);

			tile->child[1] = p;

			/*  d* -> { e00*, d10* } */
			memcpy(tile->child[0]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[1]->Fcode, tile->Fcode, 2*MAX_SUBD);

			tile->child[0]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[0]->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[1]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[1]->Fcode[2 * (tile->level - 1) + 1] = 1;

		break;

		case TILE_E:
			/* f */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_F;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v0 */
			p->v[0][0] = tile->v[0][0];
			p->v[0][1] = tile->v[0][1];

			/* p.v1 = (1 - phi^-1) * tile.v0 + (tile.v1 * phi^-1) */
			p->v[1][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][0]
				   +        GOLDEN_RATIO_1  * tile->v[1][0];
			p->v[1][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][1]
				   +        GOLDEN_RATIO_1  * tile->v[1][1];

			/* p.v2 = (1 - phi^-1) * tile.v0 + (tile.v2 * phi^-1) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][0]
				   +        GOLDEN_RATIO_1  * tile->v[2][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][1]
				   +        GOLDEN_RATIO_1  * tile->v[2][1];

			calc_tile_dir(p);

			tile->child[0] = p;
			
			/* c */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_C;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v2 */
			p->v[0][0] = tile->v[2][0];
			p->v[0][1] = tile->v[2][1];

			/* p.v1 = (1 - phi^-1) * tile.v0 + (tile.v2 * phi^-1) */
			p->v[1][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][0]
				   +        GOLDEN_RATIO_1  * tile->v[2][0];
			p->v[1][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][1]
				   +        GOLDEN_RATIO_1  * tile->v[2][1];

			/* p.v2 = (1 - phi^-1) * tile.v0 + (tile.v1 * phi^-1) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][0]
				   +        GOLDEN_RATIO_1  * tile->v[1][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][1]
				   +        GOLDEN_RATIO_1  * tile->v[1][1];

			calc_tile_dir(p);

			tile->child[1] = p;

			/* e */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_E;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v1 */
			p->v[0][0] = tile->v[1][0];
			p->v[0][1] = tile->v[1][1];

			/* p.v1 = tile.v2 */
			p->v[1][0] = tile->v[2][0];
			p->v[1][1] = tile->v[2][1];

			/* p.v2 = (1 - phi^-1) * tile.v0 + (tile.v1 * phi^-1) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][0]
				   +        GOLDEN_RATIO_1  * tile->v[1][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][1]
				   +        GOLDEN_RATIO_1  * tile->v[1][1];

			calc_tile_dir(p);

			tile->child[2] = p;

			/* a */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_A;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = (1 - phi^-1) * tile.v0 - (tile.v2 * phi^-1) */
			p->v[0][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][0]
				   +        GOLDEN_RATIO_1  * tile->v[2][0];
			p->v[0][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[0][1]
				   +        GOLDEN_RATIO_1  * tile->v[2][1];

			/* p.v1 = tile.v0 */
			p->v[1][0] = tile->v[0][0];
			p->v[1][1] = tile->v[0][1];

			calc_tile_dir(p);

			/* negate y, and swap x and y*/
			p->dir[1][0] = -p->dir[1][0];
			p->dir[1][1] = -p->dir[1][1];
			swap_dir(p->dir[0], p->dir[1]);


			tile->child[3] = p;

			/*  e* -> { f00*, c10*, e01*, a10* } */
			memcpy(tile->child[0]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[1]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[2]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[3]->Fcode, tile->Fcode, 2*MAX_SUBD);

			tile->child[0]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[0]->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[1]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[1]->Fcode[2 * (tile->level - 1) + 1] = 1;

			tile->child[2]->Fcode[2 * (tile->level - 1) + 0] = 1;
			tile->child[2]->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[3]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[3]->Fcode[2 * (tile->level - 1) + 1] = 1;
		break;

		case TILE_F:

			/* e */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_E;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = (1 - phi^-1) * tile.v1 - (tile.v0 * phi^-1) */
			p->v[0][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][0]
				   +        GOLDEN_RATIO_1  * tile->v[0][0];
			p->v[0][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][1]
				   +        GOLDEN_RATIO_1  * tile->v[0][1];

			/* p.v1 = tile.v1 */
			p->v[1][0] = tile->v[1][0];
			p->v[1][1] = tile->v[1][1];

			/* p.v2 = (1 - phi^-1) * tile.v1 + (tile.v2 * phi^-1) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][0]
				   +        GOLDEN_RATIO_1  * tile->v[2][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][1]
				   +        GOLDEN_RATIO_1  * tile->v[2][1];

			calc_tile_dir(p);

			tile->child[0] = p;
			
			/* d */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_D;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = (1 - phi^-1) * tile.v1 + (tile.v2 * phi^-1) */
			p->v[0][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][0]
				   +        GOLDEN_RATIO_1  * tile->v[2][0];
			p->v[0][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][1]
				   +        GOLDEN_RATIO_1  * tile->v[2][1];

			/* p.v1 = tile.v2 */
			p->v[1][0] = tile->v[2][0];
			p->v[1][1] = tile->v[2][1];

			/* p.v2 = (1 - phi^-1) * tile.v1 - (tile.v0 * phi^-1) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][0]
				   +        GOLDEN_RATIO_1  * tile->v[0][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][1]
				   +        GOLDEN_RATIO_1  * tile->v[0][1];

			calc_tile_dir(p);

			tile->child[1] = p;

			/* f */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_F;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = tile.v2 */
			p->v[0][0] = tile->v[2][0];
			p->v[0][1] = tile->v[2][1];

			/* p.v1 = tile.v0 */
			p->v[1][0] = tile->v[0][0];
			p->v[1][1] = tile->v[0][1];

			/* p.v2 = (1 - phi^-1) * tile.v1 - (tile.v0 * phi^-1) */
			p->v[2][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][0]
				   +        GOLDEN_RATIO_1  * tile->v[0][0];
			p->v[2][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][1]
				   +        GOLDEN_RATIO_1  * tile->v[0][1];

			calc_tile_dir(p);

			tile->child[2] = p;

			/* a */
			p = new_tile();
			p->parent = tile;
			p->type  = TILE_A;
			p->scale = tile->scale * GOLDEN_RATIO_1;
			p->level = tile->level + 1;

			/* p.v0 = (1 - phi^-1) * tile.v1 - (tile.v0 * phi^-1) */
			p->v[0][0] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][0]
				   +        GOLDEN_RATIO_1  * tile->v[0][0];
			p->v[0][1] = (1.0 - GOLDEN_RATIO_1) * tile->v[1][1]
				   +        GOLDEN_RATIO_1  * tile->v[0][1];

			/* p.v1 = tile.v0 */
			p->v[1][0] = tile->v[0][0];
			p->v[1][1] = tile->v[0][1];

			calc_tile_dir(p);

			/* negate x, and swap x and y direction */
			p->dir[0][0] = -p->dir[0][0];
			p->dir[0][1] = -p->dir[0][1];
			swap_dir(p->dir[0], p->dir[1]);

			tile->child[3] = p;

			/*  f* -> { e00*, d10*, f01*, a01* } */
			memcpy(tile->child[0]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[1]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[2]->Fcode, tile->Fcode, 2*MAX_SUBD);
			memcpy(tile->child[3]->Fcode, tile->Fcode, 2*MAX_SUBD);

			tile->child[0]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[0]->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[1]->Fcode[2 * (tile->level - 1) + 0] = 0;
			tile->child[1]->Fcode[2 * (tile->level - 1) + 1] = 1;

			tile->child[2]->Fcode[2 * (tile->level - 1) + 0] = 1;
			tile->child[2]->Fcode[2 * (tile->level - 1) + 1] = 0;

			tile->child[3]->Fcode[2 * (tile->level - 1) + 0] = 1;
			tile->child[3]->Fcode[2 * (tile->level - 1) + 1] = 0;
		break;

		default:
		break;
	}
}

int
GetMaxLOSWithinTile(tile_t *tile)
{
	int k;
	double max_imp;

	max_imp = FindMaxImportance(tile);

	if (max_imp < 1.0e-6) return 0;

	/* k = ceil( log_{gr^2} max_{tile}(mag I(x, y)) ) */
	k = (int)ceil(log(gmag * max_imp) / log(GOLDEN_RATIO * GOLDEN_RATIO));

	if (k >= MAX_SUBD) k = MAX_SUBD - 1;

	return k;
}

int
GetLocalImportance(tile_t *tile)
{
	const int n = 8.0;		/* as in the paper */
	int iv;
	int x, y;

	x = (int)(tile->v[0][0]);
	y = (int)(tile->v[0][1]);
	if (x < 0) x = 0;
	if (y < 0) y = 0;
	if (x >= MAPSIZE) x = MAPSIZE-1;
	if (y >= MAPSIZE/2) y = MAPSIZE/2-1;

	/* iv = floor( n * Psi(mag * I(x,y)) ) */
	iv = (int)floor((double)n * Psi(gmag * gimportancemap[y][x]));

	//assert(iv < 8);
	return iv;
}

void
draw_hierarcy(tile_t *t)
{
	int    i;
	int    local_LOS;
	double local_importance;

	local_LOS = GetMaxLOSWithinTile(t);

	if (t->level > local_LOS) {
		/* terminal. don't need more subdivisions */

		//draw_tile_notraverse(t);

		if (t->type == TILE_C ||
		    t->type == TILE_D ||
		    t->type == TILE_E ||
		    t->type == TILE_F) {
			/* not a "sampling tile". do nothing. */

			return;
		}

		local_importance = GetLocalImportance(t);

		if (local_importance >= FibToDecimal(t->Fcode, t->level * 2)) {

			return;	
		}

	} else {
		/* need more subdivisions */
		for (i = 0; i < 4; i++) {
			if (t->child[i]) {
				draw_hierarcy(t->child[i]);
			}
		}
	}
}

void
AdaptiveSamp(double samples[][2], double samplecvs[][2], int *nsamples, tile_t *t)
{
	int    i;
	int    is;
	int    local_LOS;
	int    iv;
	double local_importance;

	local_LOS = GetMaxLOSWithinTile(t);

	if (t->level > local_LOS) {
		/* terminal. don't need more subdivisions */

		if (t->type == TILE_C ||
		    t->type == TILE_D ||
		    t->type == TILE_E ||
		    t->type == TILE_F) {
			/* not a "sampling tile". do nothing. */

			return;
		}

		iv = GetLocalImportance(t);

		local_importance = gmag * (double)iv;

		if (local_importance >= FibToDecimal(t->Fcode, t->level * 2)) {
			/* output sample */
			is = get_structual_index(t);

			samples[(*nsamples)][0]   = t->v[0][0];
			samplecvs[(*nsamples)][0] = cv[iv][is][0];	
			samples[(*nsamples)][1]   = t->v[0][1];
			samplecvs[(*nsamples)][1] = cv[iv][is][1];	
			
			(*nsamples)++;


			return;	
		}

	} else {
		/* need more subdivisions */
		for (i = 0; i < 4; i++) {
			if (t->child[i]) {
				AdaptiveSamp(samples, samplecvs, nsamples, t->child[i]);
			}
		}
	}
}
	

void
menu(int entory)
{
	switch (entory) {
		default:
		break;
	}

	glutPostRedisplay();
}

void
reshape(int w, int h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f, (float)w / (float)h, 0.5f, 500.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	width = w; height = h;
}

void
keyboard(unsigned char k, int x, int y)
{
	int i;

	(void)x; (void)y;

	switch(k) {
	case 27:	/* ESC */
	case 'q':
		exit(0);
		break;

	case ' ':	/* space key */
		break;

	case 'j':
		showlevel++;
		if (showlevel > MAX_SUBD) showlevel = MAX_SUBD;
		break;

	case 'k':
		showlevel--;
		if (showlevel < 1) showlevel = 1;
		break;

	case 'n':
		gmag += 100.0;
		gnsamples = 0;
		for (i = 0; i < NTILES * NTILES * 4; i++) {
			AdaptiveSamp(gsamples, gsamplecvs, &gnsamples, gtiles[i]);
		}
		printf("nsamples = %d\n", gnsamples);
		break;

	case 'm':
		gmag -= 100.0;
		if (gmag < 0.0) gmag = 0.0;
		gnsamples = 0;
		for (i = 0; i < NTILES * NTILES * 4; i++) {
			AdaptiveSamp(gsamples, gsamplecvs, &gnsamples, gtiles[i]);
		}
		printf("nsamples = %d\n", gnsamples);
		break;

	case 'z':
		show_tiles = !show_tiles;
		break;

	case 'x':
		show_hierarcy = !show_hierarcy;
		break;

	default:
		break;
	}

	glutPostRedisplay();
}

void
draw_importancemap()
{
	glBindTexture(RECTANGLER_TARGET, gtexid);
	glEnable(RECTANGLER_TARGET);

	glBegin(GL_QUADS);
	//glTexCoord2i(0, MAPSIZE/2);
	glTexCoord2i(0, 0);
	glVertex2i(0, 0);
	//glTexCoord2i(MAPSIZE, MAPSIZE/2);
	glTexCoord2i(MAPSIZE, 0);
	glVertex2i(MAPSIZE, 0);
	//glTexCoord2i(MAPSIZE, 0);
	glTexCoord2i(MAPSIZE, MAPSIZE/2);
	glVertex2i(MAPSIZE, MAPSIZE/2);
	//glTexCoord2i(0, 0);
	glTexCoord2i(0, MAPSIZE/2);
	glVertex2i(0, MAPSIZE/2);
	glEnd();

	glDisable(RECTANGLER_TARGET);
}

void
draw_sampling_points()
{
	int i;
	float scale = 20.0;

	glColor3f(1.0f, 1.0f, 1.0f);
	glBegin(GL_POINTS);
	for (i = 0; i < gnsamples; i++) {
		glVertex2f(gsamples[i][0], gsamples[i][1]);
	}
	glEnd();

	glColor3f(0.0f, 1.0f, 0.0f);
	glBegin(GL_POINTS);
	for (i = 0; i < gnsamples; i++) {
		glVertex2f(gsamples[i][0] + gsamplecvs[i][0] * scale,
			   gsamples[i][1] + gsamplecvs[i][1] * scale);
	}
	glEnd();

	glColor3f(1.0, 0.0, 0.0);
	for (i = 0; i < gnsamples; i++) {
		glBegin(GL_LINES);
		glVertex2f(gsamples[i][0], gsamples[i][1]);
		glVertex2f(gsamples[i][0] + gsamplecvs[i][0] * scale,
			   gsamples[i][1] + gsamplecvs[i][1] * scale);
		glEnd();
	}

	glColor3f(1.0, 1.0, 1.0);
}

void
draw_tile_notraverse(tile_t *tile)
{
	/* draw tile */
	glColor4f(colors[tile->type][0],
		  colors[tile->type][1],
		  colors[tile->type][2],
		  0.5);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	if (tile->type != TILE_A &&
	    tile->type != TILE_B) {
		glBegin(GL_TRIANGLES);
			glVertex2f(tile->v[0][0], tile->v[0][1]);
			glVertex2f(tile->v[1][0], tile->v[1][1]);
			glVertex2f(tile->v[2][0], tile->v[2][1]);
		glEnd();
		}	

	glDisable(GL_BLEND);
}

void
draw_tile_traverse(tile_t *tile, int level)
{
	int i;
	float center[2];

	if (!tile) return;

	if (tile->level == level) {
		if (tile->type != TILE_A &&
		    tile->type != TILE_B) {

			/* draw tile */
			glColor4f(colors[tile->type][0],
				  colors[tile->type][1],
				  colors[tile->type][2],
				  0.5);
			glEnable(GL_BLEND);
			glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

			glBegin(GL_TRIANGLES);
				glVertex2f(tile->v[0][0], tile->v[0][1]);
				glVertex2f(tile->v[1][0], tile->v[1][1]);
				glVertex2f(tile->v[2][0], tile->v[2][1]);
			glEnd();

			glDisable(GL_BLEND);

			center[0] = tile->v[0][0] + tile->v[1][0] + tile->v[2][0];
			center[0] /= 3;
			center[1] = tile->v[0][1] + tile->v[1][1] + tile->v[2][1];
			center[1] /= 3;

			/* draw tile's direction */
#if 0
			glColor3f(1.0, 0.0, 0.0);
			glBegin(GL_LINES);
				//glVertex2f(tile->v[0][0], tile->v[0][1]);
				//glVertex2f(tile->v[0][0] + tile->dir[0][0] * 10.0, tile->v[0][1] + tile->dir[0][1] * 10.0);
				glVertex2f(center[0], center[1]);
				glVertex2f(center[0] + tile->dir[0][0] * 20.0, center[1] + tile->dir[0][1] * 20.0);
			glEnd();

			glColor3f(0.0, 1.0, 0.0);
			glBegin(GL_LINES);
				//glVertex2f(tile->v[0][0], tile->v[0][1]);
				//glVertex2f(tile->v[0][0] + tile->dir[1][0] * 10.0, tile->v[0][1] + tile->dir[1][1] * 10.0);
				glVertex2f(center[0], center[1]);
				glVertex2f(center[0] + tile->dir[1][0] * 20.0, center[1] + tile->dir[1][1] * 20.0);
			glEnd();
#endif
		}	


		return;
	}

	/* down to children */
	for (i = 0; i < 4; i++) {
		if (tile->child[i]) draw_tile_traverse(tile->child[i], level);
	}
}

void
draw_sampling_tiles(tile_t *tile, int level)
{
	int i;

	if (tile->level == level) {
		/* draw tile */
		glColor3fv(colors[tile->type]);

		glPointSize(5.0);
		if (tile->type == TILE_A ||
		    tile->type == TILE_B) {
			glBegin(GL_POINTS);
				glVertex2f(tile->v[0][0], tile->v[0][1]);
			glEnd();

			//print_Fcode(tile);

			glLineWidth(2.0);
			/* draw tile's direction */
			glColor3f(1.0, 0.0, 0.0);
			glBegin(GL_LINES);
				glVertex2f(tile->v[0][0], tile->v[0][1]);
				glVertex2f(tile->v[0][0] + tile->dir[0][0] * 20.0, tile->v[0][1] + tile->dir[0][1] * 20.0);
			glEnd();

			glColor3f(0.0, 1.0, 0.0);
			glBegin(GL_LINES);
				glVertex2f(tile->v[0][0], tile->v[0][1]);
				glVertex2f(tile->v[0][0] + tile->dir[1][0] * 20.0, tile->v[0][1] + tile->dir[1][1] * 20.0);
			glEnd();
			glLineWidth(1.0);
		}	

		//FindMaxImportance(tile);


		return;
	}

	/* down to children */
	for (i = 0; i < 4; i++) {
		if (tile->child[i]) draw_sampling_tiles(tile->child[i], level);
	}
}

void
draw_tile(tile_t *tile, int level)
{
	if (!tile) return;
	if (level < 0 || level >= MAX_SUBD) return;

	/* draw tiles of type C, D, E and F. */
	draw_tile_traverse(tile, level);

	/* draw pentagonal sampling tiles of type A and B */
	draw_sampling_tiles(tile, level);
}


void
display()
{
	int i;

	glClearColor(0.0, 0.0, 0.0, 0.0);
	glClear(GL_COLOR_BUFFER_BIT);

	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);

	glViewport(0.0, 0.0, WINDOW_WIDTH, WINDOW_HEIGHT);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluOrtho2D(0, gscale*2, 0, gscale);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	draw_importancemap();

	if (show_tiles) {
		for (i = 0; i < NTILES * NTILES * 4; i++) {
			draw_tile(gtiles[i], showlevel);
		}
	}

	if (show_hierarcy) {
		for (i = 0; i < NTILES * NTILES * 4; i++) {
			draw_hierarcy(gtiles[i]);
		}
	}
	draw_sampling_points();

	glutSwapBuffers();
}

void
subdivision_traverse(tile_t *tile, int level, const int maxlevel)
{
	int i;

	if (level >= maxlevel) return;
	if (!tile) return;

	subdivide(tile);

	for (i = 0; i < 4; i++) {
		if (tile->child[i]) {
			subdivision_traverse(tile->child[i],
					     level + 1,
					     maxlevel);
		}
	}
}

int
fib(int n)
{
	if (n <= 0) return 0;
	if (n < 2) return 1;

	return fib(n - 2) + fib(n - 1);
}

int
init()
{
	int i, j;
	int n;
	double scale = gscale;
	double offset[2];

	/* calcluate Fibonacci number */
	for (i = 0; i < MAX_SUBD * 2 + 3; i++) {
		FibNum[i] = fib(i);
	}

	scale  = (double)MAPSIZE / (double)NTILES;

	/* fill the plane with tile pair of 'e' and 'f' */
	for (j = 0; j < NTILES; j++) {
		for (i = 0; i < NTILES; i++) {
			n = 4 * (j * NTILES + i);		

			offset[0] = i * scale * (vertices[2] - vertices[0]);
			offset[1] = j * scale * (vertices[5] - vertices[1]) * 2;

			/* tile E */
			gtiles[n+0] = new_tile();
			gtiles[n+0]->type = TILE_E;
			gtiles[n+0]->level = 1;
			gtiles[n+0]->scale = gscale;
			gtiles[n+0]->v[0][0] = offset[0] + vertices[0] * scale;
			gtiles[n+0]->v[0][1] = offset[1] + vertices[1] * scale;
			gtiles[n+0]->v[1][0] = offset[0] + vertices[2] * scale;
			gtiles[n+0]->v[1][1] = offset[1] + vertices[3] * scale;
			gtiles[n+0]->v[2][0] = offset[0] + vertices[4] * scale;
			gtiles[n+0]->v[2][1] = offset[1] + vertices[5] * scale;
			memset(gtiles[n+0]->Fcode, 0, 2 * MAX_SUBD);
			calc_tile_dir(gtiles[n+0]);

			/* tile F */
			gtiles[n+1] = new_tile();
			gtiles[n+1]->type = TILE_F;
			gtiles[n+1]->level = 1;
			gtiles[n+1]->scale = gscale;
			gtiles[n+1]->v[0][0] = offset[0] + vertices[2] * scale;
			gtiles[n+1]->v[0][1] = offset[1] + vertices[3] * scale;
			gtiles[n+1]->v[1][0] = offset[0] + vertices[0] * scale;
			gtiles[n+1]->v[1][1] = offset[1] + vertices[1] * scale;
			gtiles[n+1]->v[2][0] = offset[0] + vertices[4] * scale;
			gtiles[n+1]->v[2][1] = offset[1] - vertices[5] * scale;
			memset(gtiles[n+1]->Fcode, 0, 2 * MAX_SUBD);
			calc_tile_dir(gtiles[n+1]);

			offset[0] = scale * vertices[2] + i * scale * (vertices[2] - vertices[0]);
			offset[1] = scale * vertices[5] + j * scale * (vertices[5] - vertices[1]) * 2;

			/* +90 tile F */
			gtiles[n+2] = new_tile();
			gtiles[n+2]->type = TILE_F;
			gtiles[n+2]->level = 1;
			gtiles[n+2]->scale = gscale;
			gtiles[n+2]->v[0][0] = offset[0] + vertices[2] * scale;
			gtiles[n+2]->v[0][1] = offset[1] + vertices[3] * scale;
			gtiles[n+2]->v[1][0] = offset[0] + vertices[0] * scale;
			gtiles[n+2]->v[1][1] = offset[1] + vertices[1] * scale;
			gtiles[n+2]->v[2][0] = offset[0] + vertices[4] * scale;
			gtiles[n+2]->v[2][1] = offset[1] - vertices[5] * scale;
			memset(gtiles[n+2]->Fcode, 0, 2 * MAX_SUBD);
			calc_tile_dir(gtiles[n+2]);

			/* +90 tile E */
			gtiles[n+3] = new_tile();
			gtiles[n+3]->type = TILE_E;
			gtiles[n+3]->level = 1;
			gtiles[n+3]->scale = gscale;
			gtiles[n+3]->v[0][0] = offset[0] + vertices[0] * scale;
			gtiles[n+3]->v[0][1] = offset[1] + vertices[1] * scale;
			gtiles[n+3]->v[1][0] = offset[0] + vertices[2] * scale;
			gtiles[n+3]->v[1][1] = offset[1] + vertices[3] * scale;
			gtiles[n+3]->v[2][0] = offset[0] + vertices[4] * scale;
			gtiles[n+3]->v[2][1] = offset[1] + vertices[5] * scale;
			memset(gtiles[n+3]->Fcode, 0, 2 * MAX_SUBD);
			calc_tile_dir(gtiles[n+3]);
		}
	}

	for (i = 0; i < NTILES * NTILES * 4; i++) {
		subdivision_traverse(gtiles[i], 1, 10);
	}
	//subdivision_traverse(gtiles[1], 1, 10);

	gnsamples = 0;
	for (i = 0; i < NTILES * NTILES * 4; i++) {
		AdaptiveSamp(gsamples, gsamplecvs, &gnsamples, gtiles[i]);
	}

	printf("nsamples = %d\n", gnsamples);

	return 0;
}

void
animation()
{
	static float tickcount = 400.0;
	float tick = tickcount;

	int i;
	gmag += tick;

	gnsamples = 0;		
	for (i = 0; i < NTILES * NTILES * 4; i++) {
		AdaptiveSamp(gsamples, gsamplecvs, &gnsamples, gtiles[i]);
	}

	printf("Gnsamples = %d\n", gnsamples);
	
	if (gnsamples > 1000) tick = -tickcount;	
	if (gnsamples < 1) tick = tickcount;

	glutPostRedisplay();

	
}

void
longtitude_latitude2xyz(double xyz[3], double uv[2])
{
	double phi, theta;

	theta = uv[1] * M_PI;
	phi   = uv[0] * 2.0 * M_PI;

	xyz[0] =  sin(theta) * cos(phi);
	xyz[1] =  cos(theta);
	xyz[2] =  sin(theta) * sin(phi);
}

void
fetch_angular_map_from_xyz(float col[3], double xyz[3], float *img)
{
	int    t, s; 
	double r;
	double norm2;
	double u, v;

	normalize3(xyz);

	r = (1.0 / M_PI) * acos(xyz[2]);
	norm2 = xyz[0] * xyz[0] + xyz[1] * xyz[1];
	if (norm2 > 1.0e-6) r /= sqrt(norm2);

	u = xyz[0] * r;
	v = xyz[1] * r;

	u = 0.5 * u + 0.5;
	v = 0.5 * v + 0.5;

	t = u * MAPSIZE;
	s = v * MAPSIZE;

	col[0] = img[3 * (s * MAPSIZE + t) + 0];
	col[1] = img[3 * (s * MAPSIZE + t) + 1];
	col[2] = img[3 * (s * MAPSIZE + t) + 2];
}

void
angular_map2longtitude_latitude(float *out, float *img)
{
	int i, j;
	double xyz[3];
	double uv[2];
	float  col[3];

	for (j = 0; j < MAPSIZE/2; j++) {
		for (i = 0; i < MAPSIZE; i++) {
			uv[0] = i / (double)MAPSIZE;
			uv[1] = j*2 / (double)(MAPSIZE);

			longtitude_latitude2xyz(xyz, uv);

			fetch_angular_map_from_xyz(col, xyz, img);

			out[3 * (j * MAPSIZE + i) + 0] = col[0];
			out[3 * (j * MAPSIZE + i) + 1] = col[1];
			out[3 * (j * MAPSIZE + i) + 2] = col[2];
		}
	}
}

int
main(int argc, char **argv)
{
	FILE             *fp;
	float            *img, *img2;
	float             col[3];
	float             max;
	float             domega, theta;
	int               i, j;
	int               mapwidth, mapheight;
	int               ret;
	rgbe_header_info  info;

	if (argc < 2) {
		printf("usage: penrose input.hdr\n");
		exit(-1);
	}

	fp = fopen(argv[1], "rb");
	if (!fp) {
		printf("can't open file [ %s ]\n", argv[1]);
		exit(-1);
	}

	ret = RGBE_ReadHeader(fp, &mapwidth, &mapheight, &info);
	if (ret != RGBE_RETURN_SUCCESS) {
		exit(-1);
	}

	if (mapwidth != MAPSIZE ||
	    mapheight != MAPSIZE) {
		printf("mapwidth %d != MAPSIZE\n", mapwidth);
	}

	img = (float *)malloc(sizeof(float) * mapwidth * mapheight * 3);
	img2 = (float *)malloc(sizeof(float) * mapwidth * (mapheight/2) * 3);
	ret = RGBE_ReadPixels_RLE(fp, img, mapwidth, mapheight);

	memset(img2, 0, sizeof(float) * mapwidth * (mapheight/2) * 3);
	/* convert AngularMap to LongtitudeLatitude coordinate */
	angular_map2longtitude_latitude(img2, img);

	/* convert RGB to Intensity */
	max = 0.0;
	for (j = 0; j < mapheight/2; j++) {
		for (i = 0; i < mapwidth; i++) {
			col[0] = img2[3 * (j * mapwidth + i) + 0];
			col[1] = img2[3 * (j * mapwidth + i) + 1];
			col[2] = img2[3 * (j * mapwidth + i) + 2];

			/* correction by solid angle */
			theta = (double)j * M_PI / (double)(mapheight/2);

			/* dw = sin(theta) d theta d phi */
			domega = sin(theta)
			       * ((double)j / (double)(mapheight/2))
			       * ((double)i / (double)mapwidth);

			gimportancemap[j][i] = CIEcol(col) * domega;
			//gimportancemap[j][i] = (float)i;

			//if (max < (float)i) max = (float)i;
			if (max < (float)(CIEcol(col) * domega))
				max = (float)(CIEcol(col) * domega);
		}
	}

	for (j = 0; j < mapheight/2; j++) {
		for (i = 0; i < mapwidth; i++) {
			gimportancemap[j][i] /= max;
			//gimportancemap[j][i] = j * 100;
		}
	}

	printf("max = %f\n", max);


	glutInit(&argc, argv);
	glutInitWindowPosition(100, 100);
	glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
	glutCreateWindow(argv[0]);

	init();

#if 1
	for (j = 0; j < (mapheight/2); j++) {
		for (i = 0; i < mapwidth ; i++) {
#if 0
			img2[3 * (j * mapwidth + i) + 0] = 3.0 * gimportancemap[j][i];
			img2[3 * (j * mapwidth + i) + 1] = 3.0 * gimportancemap[j][i];
			img2[3 * (j * mapwidth + i) + 2] = 3.0 * gimportancemap[j][i];
#endif
			if (img2[3 * (j * mapwidth + i) + 0] > 1.0) {
				img2[3 * (j * mapwidth + i) + 0] = 1.0;
			}
			if (img2[3 * (j * mapwidth + i) + 1] > 1.0) {
				img2[3 * (j * mapwidth + i) + 1] = 1.0;
			}
			if (img2[3 * (j * mapwidth + i) + 2] > 1.0) {
				img2[3 * (j * mapwidth + i) + 2] = 1.0;
			}
		}
	}
#endif

	glGenTextures(1, &gtexid);
	print_glerror();
	glBindTexture(RECTANGLER_TARGET, gtexid);
	print_glerror();
	glTexImage2D(RECTANGLER_TARGET, 0, GL_RGB,
		     mapwidth, mapheight/2, 0,
		     GL_RGB, GL_FLOAT, img2);
	print_glerror();
	glTexParameteri(RECTANGLER_TARGET, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
	glTexParameteri(RECTANGLER_TARGET, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	glTexParameteri(RECTANGLER_TARGET, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(RECTANGLER_TARGET, GL_TEXTURE_WRAP_T, GL_CLAMP);

	print_glerror();
	
	glutReshapeFunc(reshape);
	glutDisplayFunc(display);
	glutKeyboardFunc(keyboard);
	//glutIdleFunc(animation);

	glutMainLoop();

	return 0;
}

